/*
 * Copyright (c) 2012, Enhanced Four
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of 'Enhanced Four' nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package net.indiespot.continuations;

import java.util.Arrays;
import java.util.Comparator;
import java.util.PriorityQueue;
import java.util.Queue;

import craterstudio.data.CircularArrayList;

public class VirtualProcessor {

	static final ThreadLocal<VirtualProcessor> LAST_DEFINED = new ThreadLocal<VirtualProcessor>();

	private final Queue<VirtualThread> taskQueue;
	private final CircularArrayList<VirtualThread> scheduleNextTick, executeFirstNextTick, executeFirstCurrTick;

	public VirtualProcessor() {
		this.taskQueue = new PriorityQueue<VirtualThread>(11, new VirtualThreadComparator());
		this.scheduleNextTick = new CircularArrayList<VirtualThread>();

		this.uncaughtExceptionHandler = null;

		this.executeFirstCurrTick = new CircularArrayList<VirtualThread>();
		this.executeFirstNextTick = new CircularArrayList<VirtualThread>();

		LAST_DEFINED.set(this);
	}

	private long currentTickTimestamp;

	public final long getCurrentTime() {
		return this.currentTickTimestamp;
	}

	//

	private VirtualUncaughtExceptionHandler uncaughtExceptionHandler;

	public void setUncaughtExceptionHandler(VirtualUncaughtExceptionHandler uncaughtExceptionHandler) {
		this.uncaughtExceptionHandler = uncaughtExceptionHandler;
	}

	//

	private VirtualThread current;

	/**
	 * Executes the virtual threads that are scheduled, if any. The new threads,
	 * or threads that are resumed or reschedule themselves to be executes will
	 * be executed in the next tick.
	 */

	public int tick(long now) {

		this.currentTickTimestamp = now;

		{
			// N.B.: this is super fast due to CircularArrayList

			while (!this.executeFirstNextTick.isEmpty()) {
				this.executeFirstCurrTick.addLast(this.executeFirstNextTick.removeFirst());
			}

			while (!this.scheduleNextTick.isEmpty()) {
				this.taskQueue.add(this.scheduleNextTick.removeFirst());
			}
		}

		int counter = 0;

		while (true) {
			VirtualThread thread;

			if (!this.executeFirstCurrTick.isEmpty()) {
				thread = this.executeFirstCurrTick.removeFirst();
			} else {
				thread = this.taskQueue.peek();
				if (thread == null || thread.wakeUpAt > now) {
					break;
				}

				thread = this.taskQueue.poll();
				if (thread.getState() == VirtualThreadState.SLEEPING) {
					thread.wake();
				}
			}

			this.current = thread;
			VirtualThreadState state;
			try {
				state = thread.step();
			} catch (Throwable uncaught) {
				cleanupVirtualStackTrace(uncaught);

				if (this.uncaughtExceptionHandler == null) {
					String msg = VirtualProcessor.class.getSimpleName() + " encountered uncaught exception in " + thread.getName();
					new Throwable(msg, uncaught).printStackTrace();
				} else {
					this.uncaughtExceptionHandler.uncaughtException(thread, uncaught);
				}
				continue;
			} finally {
				this.current = null;
				counter++;
			}

			switch (state) {
				case SLEEPING:
				case RUNNABLE:
					this.scheduleNextTick.addLast(thread);
					break;

				case YIELDED:
					this.executeFirstNextTick.addLast(thread);
					break;

				case SUSPENDED:
				case NEW:
				case TERMINATED:
					break;
			}
		}

		return counter;
	}

	private static void cleanupVirtualStackTrace(Throwable problem) {
		StackTraceElement[] trace = problem.getStackTrace();

		int cutoff = 0;
		for (StackTraceElement elem : trace) {
			if (elem.getClassName().equals(VirtualThread.class.getName())) {
				if (elem.getMethodName().equals("run")) {
					break;
				}
			}
			cutoff++;
		}

		trace = Arrays.copyOf(trace, cutoff);
		problem.setStackTrace(trace);
	}

	public boolean hasPendingTasks() {
		return !taskQueue.isEmpty() || !scheduleNextTick.isEmpty() || //
		   !executeFirstCurrTick.isEmpty() || !executeFirstNextTick.isEmpty();
	}

	boolean unschedule(VirtualThread thread) {
		if (this.current == thread) {
			throw new IllegalStateException("virtual thread cannot unschedule itself, use stop()");
		}
		if (this.executeFirstNextTick.remove(thread)) {
			return true;
		}
		if (this.scheduleNextTick.remove(thread)) {
			return true;
		}
		if (this.taskQueue.remove(thread)) {
			return true;
		}
		throw new IllegalStateException("virtual thread was not scheduled");
	}

	void schedule(VirtualThread thread) {
		if (thread.wakeUpAt == VirtualThread.WAKEUP_IMMEDIATELY) {
			this.executeFirstNextTick.addLast(thread);
		} else {
			this.scheduleNextTick.addLast(thread);
		}
	}

	static class VirtualThreadComparator implements Comparator<VirtualThread> {
		@Override
		public int compare(VirtualThread o1, VirtualThread o2) {
			return o1.wakeUpAt < o2.wakeUpAt ? -1 : +1;
		}
	}
}
